from ctypes import Array as ctypes_Array
from _ctypes import _CData
from collections.abc import MutableMapping
from numbers import Real
from threading import Lock
from typing import List, Optional, Protocol, Tuple, Union

from ._util.wintypes_extra import WSAPOLLFD

POLLERR: int
POLLHUP: int
POLLIN: int
POLLNVAL: int
POLLOUT: int
POLLPRI: int
POLLRDBAND: int
POLLRDNORM: int
POLLWRBAND: int
POLLWRNORM: int

class _Fileobj(Protocol):
    def fileno(self) -> int: ...

class wsapoll:
    __slots__ = ['__impl', '__fd_to_key', '__buffer']
    _registered: MutableMapping[int, int]
    __impl: ctypes_Array[WSAPOLLFD]
    __impl_uptodate: bool
    __buffer: _CData
    __lock: Lock
    def __init__(self, sizehint: int=...) -> None: ...
    def __repr__(self) -> str: ...

    def register(self, fd: Union[_Fileobj, int], eventmask: int=...) -> None: ...
    def modify(self, fd: Union[_Fileobj, int], eventmask: int) -> None: ...
    def unregister(self, fd: Union[_Fileobj, int]) -> None: ...
    def poll(self, timeout: Optional[Real]=None) -> List[Tuple[int, int]]: ...

    def _poll(self, timeout: int=-1) -> List[Tuple[int, int]]: ...
    def _clear(self) -> None: ...
    def __check_maybe_affected(self) -> bool: ...
    def __update_impl(self) -> None: ...



import ctypes
from ctypes import wintypes
from ctypes.wintypes import DWORD, SHORT

from .misc import POLL_FLAGS_FOR_REPR, repr_flags

__all__ = [
    'DWORD_PTR',
    'INVALID_SOCKET',
    'SOCKET',
    'SOCKET_ERROR',
    'UINT_PTR',
    'WSAEINTR',
    'WSAEINVAL',
    'WSAPOLLFD',
]


try: UINT_PTR = wintypes.UINT_PTR
except AttributeError: UINT_PTR = ctypes.c_size_t

try: DWORD_PTR = wintypes.DWORD_PTR
except AttributeError: DWORD_PTR = max([DWORD, UINT_PTR], key=ctypes.sizeof)


# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L139
try: SOCKET = wintypes.SOCKET
except AttributeError: SOCKET = UINT_PTR

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L166
try: INVALID_SOCKET = wintypes.INVALID_SOCKET
except AttributeError: INVALID_SOCKET = SOCKET(~0).value

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L167
try: SOCKET_ERROR = wintypes.SOCKET_ERROR
except AttributeError: SOCKET_ERROR = -1


# https://learn.microsoft.com/en-us/windows/win32/winsock/windows-sockets-error-codes-2#WSAEINTR
try: WSAEINTR = wintypes.WSAEINTR
except AttributeError: WSAEINTR = 10004

# https://learn.microsoft.com/en-us/windows/win32/winsock/windows-sockets-error-codes-2#WSAEINVAL
try: WSAEINVAL = wintypes.WSAEINVAL
except AttributeError: WSAEINVAL = 10022


_SSOCKET = {
    ctypes.c_uint64: ctypes.c_int64,
    ctypes.c_int64:  ctypes.c_int64,
    ctypes.c_uint32: ctypes.c_int32,
    ctypes.c_int32:  ctypes.c_int32,
    ctypes.c_uint16: ctypes.c_int16,
    ctypes.c_int16:  ctypes.c_int16,
    ctypes.c_uint8:  ctypes.c_int8,
    ctypes.c_int8:   ctypes.c_int8,
}[SOCKET]


class WSAPOLLFD(ctypes.Structure):
    __slots__ = []
    class _field_1(ctypes.Union):
        __slots__ = []
        _fields_ = [
            ('fd', SOCKET), # regular access in accordance with the "official" struct definition
            ('_fd', _SSOCKET), # signed access to reliably determine which fields "are negative" thus WSAPoll will ignore
        ]
    _anonymous_ = ['_1']
    _fields_ = [
        ('_1', _field_1),
        ('events', SHORT),
        ('revents', SHORT),
    ]

    def __init__(self, fd=INVALID_SOCKET, events=0, revents=0):
        self.fd = fd
        self.events = events
        self.revents = revents

    def __repr__(self):
        if self.fd != INVALID_SOCKET:
            return f"<{__name__}.{self.__class__.__name__} fd={self.fd}, events={repr_flags(self.events, POLL_FLAGS_FOR_REPR)}, revents={repr_flags(self.revents, POLL_FLAGS_FOR_REPR)}>"
        else:
            return f"<{__name__}.{self.__class__.__name__} fd=INVALID_SOCKET, ...>"



import ctypes
from ctypes import POINTER, c_void_p
from ctypes import oledll, wintypes
from ctypes.wintypes import DWORD, WORD, LPVOID
from functools import cache

from .wintypes_extra import DWORD_PTR

__all__ = [
    'getpagesize',
    'getallocationgranularity',
    'GetSystemInfo',
]


@cache
def getpagesize():
    return GetSystemInfo()['pageSize']

@cache
def getallocationgranularity():
    return GetSystemInfo()['allocationGranularity']


class SYSTEM_INFO(ctypes.Structure):
    class _field_1(ctypes.Union):
        class _field_2(ctypes.Structure):
            _fields_ = [
                ('wProcessorArchitecture', WORD),
                ('wReserved', WORD),
            ]
        _anonymous_ = ['_2']
        _fields_ = [
            ('dwOemId', DWORD),
            ('_2', _field_2),
        ]
    _anonymous_ = ['_1']
    _fields_ = [
        ('_1', _field_1),
        ('dwPageSize', DWORD),
        ('lpMinimumApplicationAddress', LPVOID),
        ('lpMaximumApplicationAddress', LPVOID),
        ('dwActiveProcessorMask', DWORD_PTR),
        ('dwNumberOfProcessors', DWORD),
        ('dwProcessorType', DWORD),
        ('dwAllocationGranularity', DWORD),
        ('wProcessorLevel', WORD),
        ('wProcessorRevision', WORD),
    ]

    @property
    def value(self):
        return {
            'pageSize': self.dwPageSize,
            'minimumApplicationAddress': c_void_p(self.lpMinimumApplicationAddress),
            'maximumApplicationAddress': c_void_p(self.lpMaximumApplicationAddress),
            'activeProcessorMask': self.dwActiveProcessorMask,
            '_activeProcessors': frozenset((lambda mask, n: (i for i in range(n) if (mask & (1<<i))))(self.dwActiveProcessorMask, self.dwNumberOfProcessors)),
            'numberOfProcessors': self.dwNumberOfProcessors,
            'allocationGranularity': self.dwAllocationGranularity,
            'processorLevel': self.wProcessorLevel,
            'processorRevision': self.wProcessorRevision,
            'processorArchitecture': self.wProcessorArchitecture,
            'oemId': self.dwOemId
        }


_GetNativeSystemInfo = oledll.Kernel32['GetNativeSystemInfo']

_GetNativeSystemInfo.argtypes = [
    POINTER(SYSTEM_INFO),
]


def GetSystemInfo():
    result = SYSTEM_INFO()
    _GetNativeSystemInfo(result)
    return result.value



from ctypes import c_void_p
from typing import Set, TypedDict

class _SystemInfo(TypedDict):
    pageSize: int
    minimumApplicationAddress: c_void_p
    maximumApplicationAddress: c_void_p
    activeProcessorMask: int
    _activeProcessors: Set[int]
    numberOfProcessors: int
    allocationGranularity: int
    processorLevel: int
    processorRevision: int
    processorArchitecture: int
    oemId: int

def getpagesize() -> int: ...
def getallocationgranularity() -> int: ...
def GetSystemInfo() -> _SystemInfo: ...



import select

__all__ = [
    'POLLERR',
    'POLLHUP',
    'POLLIN',
    'POLLNVAL',
    'POLLOUT',
    'POLLPRI',
    'POLLRDBAND',
    'POLLRDNORM',
    'POLLWRBAND',
    'POLLWRNORM',
]


# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L325
try: POLLERR = select.POLLERR
except AttributeError: POLLERR = 0x0001

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L326
try: POLLHUP = select.POLLHUP
except AttributeError: POLLHUP = 0x0002

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L327
try: POLLNVAL = select.POLLNVAL
except AttributeError: POLLNVAL = 0x0004

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L328
try: POLLWRNORM = select.POLLWRNORM
except AttributeError: POLLWRNORM = 0x0010

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L329
try: POLLWRBAND = select.POLLWRBAND
except AttributeError: POLLWRBAND = 0x0020

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L330
try: POLLRDNORM = select.POLLRDNORM
except AttributeError: POLLRDNORM = 0x0100

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L331
try: POLLRDBAND = select.POLLRDBAND
except AttributeError: POLLRDBAND = 0x0200

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L332
try: POLLPRI = select.POLLPRI
except AttributeError: POLLPRI = 0x0400

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L333
try: POLLIN = select.POLLIN
except AttributeError: POLLIN = (POLLRDNORM | POLLRDBAND)

# https://github.com/wine-mirror/wine/blob/wine-10.12/include/winsock2.h#L334
try: POLLOUT = select.POLLOUT
except AttributeError: POLLOUT = POLLWRNORM



from contextlib import contextmanager
from math import ceil, floor
from operator import index

from .select_extra import *

__all__ = [
    'POLL_FLAGS_FOR_REPR',
    'enter_or_die',
    'getfd',
    'repr_flags',
    'smallest_multiple_atleast',
    'uptruncate',
]


POLL_FLAGS_FOR_REPR = {
    'POLLIN': POLLIN, # prefer reporting semantic POLLIN over its components, when applicable
    'POLLRDNORM': POLLRDNORM,
    'POLLRDBAND': POLLRDBAND,
    'POLLPRI': POLLPRI,
    'POLLOUT': POLLOUT, # prefer reporting semantic POLLOUT over its components, when applicable
    'POLLWRNORM': POLLWRNORM,
    'POLLWRBAND': POLLWRBAND,
    'POLLERR': POLLERR,
    'POLLHUP': POLLHUP,
    'POLLNVAL': POLLNVAL,
}


@contextmanager
def enter_or_die(lock, error_or_message):
    if lock.acquire(blocking=False):
        try:
            yield lock
        finally:
            lock.release()
    else:
        if isinstance(error_or_message, BaseException) or issubclass(error_or_message, BaseException):
            raise error_or_message
        else:
            raise RuntimeError(error_or_message)


def getfd(fileobj):
    return int(fileobj.fileno()) if hasattr(fileobj, 'fileno') else index(fileobj)


def uptruncate(x):
    "cast float *x* to an integer, rounding away from zero."
    if not (x < 0.0):
        return ceil(x)
    else:
        return floor(x)


def smallest_multiple_atleast(base, minimum_value):
    "return the smallest multiple of *base* less than *minimum_value*."
    return base * ((minimum_value + base - 1) // base)


def repr_flags(mask, flags):
    acc1 = int(mask)
    acc2 = []

    for name, value in flags.items():
        if (acc1 & value) == value:
            acc1 &= ~value
            acc2.append(name)

    if acc1 != 0 or len(acc2) == 0:
        acc2.append(str(acc1))
    
    if len(acc2) == 1:
        return acc2[0]

    return f"({' | '.join(acc2)})"



from ctypes import WinError, byref, create_string_buffer, memmove, resize, sizeof
from ctypes import windll
from ctypes.wintypes import INT, LPVOID, ULONG
from errno import ENOENT
import logging
from socket import SOCK_STREAM, socket as socket_
import sys
from threading import Lock
from time import monotonic_ns

from ._util.select_extra import *
from ._util import (
    POLL_FLAGS_FOR_REPR,
    SOCKET_ERROR,
    WSAEINTR,
    WSAPOLLFD,
    enter_or_die,
    getallocationgranularity,
    getfd,
    repr_flags,
    smallest_multiple_atleast,
    uptruncate,
)

__all__ = [
    'POLLERR',
    'POLLHUP',
    'POLLIN',
    'POLLNVAL',
    'POLLOUT',
    'POLLPRI',
    'POLLRDBAND',
    'POLLRDNORM',
    'POLLWRBAND',
    'POLLWRNORM',
    'wsapoll',
]

IS_PRE_19041 = sys.getwindowsversion() < (10, 0, 19041)
_POLL_DISCONNECTION = POLLHUP | POLLERR | POLLWRNORM


_WSAPoll = windll.Ws2_32['WSAPoll']

_WSAPoll.argtypes = [
    LPVOID,
    ULONG,
    INT,
]


_WSAGetLastError = windll.Ws2_32['WSAGetLastError']


class wsapoll:
    __slots__ = [
        '_registered',
        '__impl',
        '__impl_uptodate',
        # We have to track the buffer separately to avoid freaking ctypes out
        # if resize is called more than once; only the originally allocated
        # object "owns" the memory, even after a call to resize. There is no way
        # to robustly resize ctypes.Array instances at this time, so we are
        # just keeping the original buffer around, in addition to impl, which
        # is a subordinate "view" of only the buffer's allocated slots.
        # https://github.com/python/cpython/issues/65527
        # https://docs.python.org/3/library/ctypes.html#ctypes._CData._b_needsfree_
        '__buffer',
        # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L661-L666
        # https://github.com/pypy/pypy/blob/release-pypy3.11-v7.3.18/pypy/module/select/interp_select.py#L87-L88
        '__lock',
    ]

    def __init__(self, sizehint=max(getallocationgranularity() // sizeof(WSAPOLLFD), 1)):
        self._registered = {}
        impl_t = WSAPOLLFD * 0
        self.__buffer = buf = (impl_t._type_ * sizehint)()
        self.__impl = impl_t.from_buffer(buf)
        self.__impl_uptodate = True
        self.__lock = Lock()

    def __repr__(self):
        return f"<{__name__}.{self.__class__.__name__} {self._registered!r}>"

    def __check_maybe_affected(self):
        return any(
            (
                fd >= 0
                and eventmask == _POLL_DISCONNECTION
            )
            for fd, eventmask in self._registered.items()
        )

    def poll(self, timeout=None):
        with enter_or_die(self.__lock, "concurrent poll() invocation"):
            if (not IS_PRE_19041) and (timeout is None) and self.__check_maybe_affected():
                logging.warning("Outbound TCP connection failures won't be reported by wsapoll.poll() on versions of Windows prior to \"Windows 10 version 2004 (OS build 19041)\"; consider updating the operating system, using IOCP (via asyncio), or setting a finite timeout.\nFor more information, see https://daniel.haxx.se/blog/2012/10/10/wsapoll-is-broken/")

            if not self.__impl_uptodate:
                self.__update_impl()

            timeout_ms = uptruncate(timeout * 1000) if timeout is not None else -1
            return self._poll(timeout_ms)

    def _poll(self, timeout=-1):
        impl = self.__impl
        impl_len = len(impl)

        # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L661-L666
        # FIXME: raise if called concurrently on the same thread

        # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L645-L647
        if timeout >= 0:
            timeout_deadline = monotonic_ns() // 1000 + timeout

        # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L675-L701
        while True:
            # no need to call "byref" as that's already how ctypes handles arrays passed as LPVOID
            ret = _WSAPoll(impl, impl_len, timeout)

            # https://learn.microsoft.com/en-us/windows/win32/api/winsock2/nf-winsock2-wsapoll#return-value
            if ret == SOCKET_ERROR:
                errno = _WSAGetLastError()

                # https://peps.python.org/pep-0475/
                if errno == WSAEINTR:
                    # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L692-L699
                    if timeout >= 0:
                        timeout = max(timeout_deadline - monotonic_ns() // 1000, 0)
                    continue

                raise WinError(errno)

            assert 0 <= ret <= impl_len
            break

        return [
            (fd, events)
            for fd, events in ((slot.fd, slot.revents) for slot in impl)
                if events
        ]

    def __update_impl(self):
        registered = self._registered
        impl = self.__impl
        buf = self.__buffer

        fds = len(registered)
        impl_t = impl._type_ * fds

        if sizeof(impl_t) > sizeof(buf):
            # ...But first, actually purchase moar RAM
            resize(
                buf,
                smallest_multiple_atleast(
                    getallocationgranularity(),
                    max(
                        sizeof(impl._type_ * (len(impl) * 2)),
                        sizeof(impl_t)
                    )
                )
            )

        self.__impl = impl = impl_t.from_buffer(self.__buffer)

        for slot, (fd, eventmask) in zip(impl, registered.items()):
            slot.fd = fd
            slot.events = eventmask

        self.__impl_uptodate = True

    def register(self, fileobj, eventmask=(POLLIN | POLLPRI | POLLOUT)):
        fd = getfd(fileobj)
        with self.__lock:
            self._registered[fd] = eventmask

            self.__impl_uptodate = False

    def unregister(self, fileobj):
        fd = getfd(fileobj)
        with self.__lock:
            try:
                del self._registered[fd]
            except KeyError:
                raise KeyError(f"{fileobj!r} is not registered") from None

            self.__impl_uptodate = False

    def modify(self, fileobj, eventmask):
        fd = getfd(fileobj)
        with self.__lock:
            try:
                self._registered[fd]
            except KeyError:
                # https://github.com/python/cpython/blob/v3.13.0/Modules/selectmodule.c#L548
                raise OSError(ENOENT, f"{fileobj!r} is not registered") from None
            else:
                self._registered[fd] = eventmask

            self.__impl_uptodate = False

    def _clear(self):
        with self.__lock:
            self._registered.clear()
            self.__update_impl()

    def __getstate__(self):
        return self._registered

    def __setstate__(self, state):
        self.__init__(sizehint=len(state))
        self._registered.update(state)
        self.__update_impl()



from contextlib import contextmanager
from numbers import Real
from threading import Lock
from typing import Iterator, Mapping, Protocol, Union

class _Fileobj(Protocol):
    def fileno(self) -> int: ...

POLL_FLAGS_FOR_REPR: Mapping[str, int]

@contextmanager
def enter_or_die(lock: Lock) -> Iterator[Lock]: yield
def getfd(fd: Union[_Fileobj, int]) -> int: ...
def repr_flags(mask: int, flags: Mapping[str, int]) -> str: ...
def smallest_multiple_atleast(base: int, minimum_value: int) -> int: ...
def uptruncate(x: Real) -> int: ...
